#!/usr/local/sbin/charm-env python3
"""Juju actions to operate Graylog."""
import datetime
import os
import sys
import traceback

hooks_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "hooks"))
if hooks_dir not in sys.path:
    sys.path.append(hooks_dir)
libs_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "lib"))
if libs_dir not in sys.path:
    sys.path.append(libs_dir)

from charmhelpers.core import hookenv, host, unitdata  # noqa: E402

from charms import reactive  # noqa: E402
from charms.layer.graylog.api import get_ignore_indexer_failures_file  # noqa: E402
from charms.layer.graylog.snap_change import (  # noqa: E402
    ChannelChangeStatus,
    get_channel_change_status,
    perform_snap_channel_change,
)


hooks = hookenv.Hooks()


def reactive_remove_state(state):
    """Remove a reactive charm state and run the reactive charm script."""
    reactive.remove_state(state)
    reactive.main()


@hooks.hook("show-admin-password")
def show_admin_password():
    """Return the stored password or generates a new one."""
    db = unitdata.kv()
    admin_password = db.get("admin_password")
    if not admin_password:
        admin_password = host.pwgen(18)
        db.set("admin_password", admin_password)
        reactive_remove_state("graylog.configured")
    hookenv.action_set({"admin-password": admin_password})


@hooks.hook("set-admin-password")
def set_admin_password():
    """Update Graylog's admin password passed by Juju action parameter."""
    admin_password = hookenv.action_get()["password"]
    db = unitdata.kv()
    stored_password = db.get("admin_password")
    if admin_password == stored_password:
        status = "no-change"
    else:
        if admin_password and admin_password != "None":
            status = "updated"
        else:
            status = "generated"
            admin_password = host.pwgen(18)
        db.set("admin_password", admin_password)

    hookenv.action_set({"admin-password": admin_password, "change": status})
    reactive_remove_state("graylog.configured")


@hooks.hook("ignore-indexer-failures")
def ignore_indexer_failures():
    """Update the indexer failures file to skip alerts about them."""
    with open(get_ignore_indexer_failures_file(), "w") as f:
        f.write(datetime.datetime.utcnow().isoformat())


@hooks.hook("do-upgrade")
def do_upgrade():
    """Verify and execute a Graylog snap upgrade."""
    status = get_channel_change_status()
    if status == ChannelChangeStatus.UPGRADE_REQUESTED:
        new_channel = hookenv.config("channel")
        perform_snap_channel_change(new_channel)
        # Let reactive hooks react to the above.
        reactive.main()
    else:
        hookenv.action_fail(
            "No action taken; app is not blocked by a pending channel upgrade"
        )


def main(argv):  # noqa: D103
    try:
        hooks.execute(argv)
    except Exception:
        hookenv.action_fail("Unhandled exception")
        tb = traceback.format_exc()
        hookenv.action_set(dict(traceback=tb))
        hookenv.log("Unhandled exception while running: {!r}".format(argv))
        print(tb)


if __name__ == "__main__":
    main(sys.argv)
