"""Helper functions to manage snaps."""
import enum
import os
import shutil

from charmhelpers.core import hookenv, host, unitdata

from charms.layer.graylog.constants import (
    CONF_FILE,
    SERVICE_NAME,
    SNAP_CONF_FILE,
    SNAP_NAME,
)
from charms.reactive import remove_state
from charms.reactive.helpers import data_changed

# Allow mocking of other layers during unit tests
try:
    from charms.layer import snap
except ImportError:
    snap = None


class ChannelChangeStatus(enum.Enum):
    """Snap channel states when performing upgrades."""

    NO_CHANGE = 1
    INVALID = 2
    UPGRADE_REQUESTED = 3
    DOWNGRADE_REQUESTED = 4
    LATEST_REQUESTED = 5
    MINOR_CHANGE = 6


def get_channel_change_status():
    """Given the current and requested channel, computes the channel change status.

    Specifically, this refers to whether there is any meaningful delta between the
    installed snap's channel and the requested channel - specifically, the track/risk,
    excluding the optional branch.

    Returns a value from the ChannelChangeStatus enum.

    """
    new_channel = hookenv.config("channel")
    cur_channel = snap.get_installed_channel(SNAP_NAME)
    if (not new_channel) or new_channel == cur_channel:
        return ChannelChangeStatus.NO_CHANGE
    elif not _is_channel_valid(new_channel):
        return ChannelChangeStatus.INVALID
    else:
        cur_track = cur_channel.split("/")[0]
        new_track = new_channel.split("/")[0]
        if new_track == "latest":
            return ChannelChangeStatus.LATEST_REQUESTED
        else:
            cur_track = int(cur_track)
            new_track = int(new_track)
            if new_track > cur_track:
                return ChannelChangeStatus.UPGRADE_REQUESTED
            elif new_track < cur_track:
                return ChannelChangeStatus.DOWNGRADE_REQUESTED
            else:
                return ChannelChangeStatus.MINOR_CHANGE


def _is_channel_valid(channel):
    """Validate snap tracks and risks."""
    try:
        track, risk = channel.split("/")[:2]
    except ValueError:
        return False
    # NOTE: While we won't allow changing to the "latest" channel, it is a valid
    # channel for the snap.  We'll reject it elsewhere.
    snap_tracks = ("2", "3", "latest")
    snap_risks = ("stable", "candidate", "beta", "edge")
    return track in snap_tracks and risk in snap_risks


def perform_snap_channel_change(new_channel):
    """Execute snap channel change, after validation passed."""
    host.service_stop(SERVICE_NAME)
    # backup config file using the current version as a suffix
    cur_snap = snap.get_installed_version(SNAP_NAME)
    if os.path.exists(CONF_FILE):
        shutil.copy2(CONF_FILE, "{path}.{ext}".format(path=CONF_FILE, ext=cur_snap))
    hookenv.status_set("maintenance", "Refreshing graylog snap {}".format(new_channel))
    snap.refresh(SNAP_NAME, channel=new_channel)
    new_snap = snap.get_installed_version(SNAP_NAME)

    # Clear data related to the upgrade block
    unitdata.kv().unset("channel_block_type")

    # When the snap version changes, use the config file from the snap and
    # adjust states to ensure appropriate re-config happens.
    if cur_snap != new_snap:
        shutil.copy2(SNAP_CONF_FILE, CONF_FILE)
        data_changed("elasticsearch.relation", None)
        data_changed("mongodb.uri", None)
        remove_state("graylog.configured")
