"""Helper functions to manage Graylog."""
from urllib.parse import urlparse

from charmhelpers.core import hookenv

from charms.layer.graylog.constants import SNAP_NAME

import netifaces


# Allow mocking of other layers during unit tests
try:
    from charms.layer import snap
except ImportError:
    snap = None


def get_api_port():  # noqa: D103
    return "9001" if is_v2() else "9000"


def get_api_url():  # noqa: D103
    return "http://127.0.0.1:{}/api/".format(get_api_port())


def is_v2():  # noqa: D103
    version = snap.get_installed_version(SNAP_NAME)
    # If the snap isn't installed yet, base our version off the charm config
    if not version:
        version = hookenv.config("channel")
    return True if version.startswith("2") else False


def validate_api_uri(uri):
    """Ensure the given API URI is valid based on the Graylog version.

    For v2, api-related options should end with '/api/'. In v3, the graylog
    server hard codes '/api/' to the end of these options and should therefore
    be removed if included in the charm config.
    """
    if not uri.endswith("/"):
        uri += "/"

    if is_v2():
        if not uri.endswith("/api/"):
            hookenv.log("Appending 'api/' to the configured REST API options")
            uri += "api/"
    else:
        if uri.endswith("/api/"):
            hookenv.log("Removing 'api/' from the configurted REST API options")
            uri = uri[:-4]

    parsed_url = urlparse(uri)
    if parsed_url.hostname == "0.0.0.0":
        uri = parsed_url._replace(
            netloc="{}:{}".format(get_ip_address(), parsed_url.port)
        ).geturl()

    return uri


def get_ip_address():
    """Determine the ip address that would be used for outgoing connection.

    This function tries to determine the network ip address, that would be used
    when access servers in the internet. It uses a dummy UDP connection todo
    so.
    As the connection itself is not made, it is safe to use any address.
    """
    # Get the default gateway's interface on this host
    # ref: https://pypi.org/project/netifaces/
    # returns an IP, interface tuple
    iface = netifaces.gateways()["default"][netifaces.AF_INET][1]

    # That will return a one-element list with following format
    # get it with [0]
    ip = netifaces.ifaddresses(iface)[netifaces.AF_INET][0]["addr"]

    hookenv.log("Found ip {} to use".format(ip))
    return ip
