"""Graylog reactive script."""
import hashlib
import os
import re
import time
from urllib.parse import urlparse

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv, host, unitdata

import charms.leadership
from charms.layer import snap
from charms.layer.graylog import (
    GraylogApi,
    LogExtractPipeline,
    get_api_port,
    get_api_url,
    is_v2,
    validate_api_uri,
)
from charms.layer.graylog.constants import (
    CONF_FILE,
    DEFAULT_REST_API_TIMEOUT,
    ELASTICSEARCH_DISCOVERY_PORT,
    NAGIOS_USERNAME,
    SERVER_DEFAULT_CONF_FILE,
    SERVICE_NAME,
    SHIPPED_SNAP_SERVER_DEFAULT_CONF_FILE,
    SNAP_NAME,
)
from charms.layer.graylog.snap_change import (
    ChannelChangeStatus,
    get_channel_change_status,
    perform_snap_channel_change,
)
from charms.reactive import (
    hook,
    is_state,
    remove_state,
    set_state,
    when,
    when_any,
    when_not,
    when_not_all,
)
from charms.reactive.helpers import data_changed

import yaml


class ApiTimeout(Exception):
    """Unable to restart Graylog in a timely manner."""

    pass


@hook("upgrade-charm")
def upgrade_charm():
    """Reconfigure Graylog upon Juju charm upgrade."""
    remove_state("graylog.configured")


@when("config.changed")
@when_not("config.changed.channel")
def update_config():
    """Allow new runtime config options to be processed.

    Reset flags to allow new runtime config options to be written. Deployment
    config options (e.g. 'channel') are handled separately.
    """
    remove_state("graylog.configured")


rotation_strategies = {
    "time": {
        "class": "org.graylog2.indexer.rotation.strategies.TimeBasedRotationStrategy",
        "special": "rotation_period",
        "juju-conf": "index_rotation_period",
    },
    "size": {
        "class": "org.graylog2.indexer.rotation.strategies.SizeBasedRotationStrategy",
        "special": "max_size",
        "juju-conf": "index_rotation_size",
    },
    "msg_count": {
        "class": (
            "org.graylog2.indexer.rotation.strategies.MessageCountRotationStrategy"
        ),
        "special": "max_docs_per_index",
        "juju-conf": "index_rotation_msg_count",
    },
}


@when_not("snap.installed.graylog")
def install_graylog():
    """Install the graylog snap.

    The snap layer will first try to install a valid 'graylog.snap' resource.
    If a valid resource is not found, the snap will be installed from the snap
    store according to the configured 'channel' option.
    """
    channel = hookenv.config("channel")

    hookenv.status_set("maintenance", "Installing graylog snap")
    snap.install(SNAP_NAME, channel=channel)


@when("snap.installed.graylog")
@when("config.changed.channel")
def refresh_graylog():
    """Refresh the graylog snap.

    A change to the 'channel' config option may warrant a snap refresh. If so,
    stop the service, backup the current config, and refresh.
    """
    new_channel = hookenv.config("channel")
    if new_channel:
        if get_channel_change_status() == ChannelChangeStatus.MINOR_CHANGE:
            perform_snap_channel_change(new_channel)
    else:
        hookenv.log("Cannot refresh snap when the 'channel' config option is missing")
    report_status()


@when("leadership.set.master")  # noqa: C901
@when("snap.installed.graylog")
@when_not("graylog.configured")
def configure_graylog():  # noqa: C901
    """Graylog configuration.

    Need to check that we have a non-empty config file, as the snap
    sometimes takes a while to set it up (after reporting it installed)
    see LP#1764394
    """
    if not (os.path.isfile(CONF_FILE) and os.path.getsize(CONF_FILE) > 0):
        hookenv.status_set("blocked", "Waiting for {}".format(CONF_FILE))
        return

    db = unitdata.kv()
    conf = hookenv.config()

    admin_password = db.get("admin_password")
    password_secret = db.get("password_secret")
    if not (admin_password and password_secret):
        hookenv.status_set("maintenance", "Waiting for leader passwords")
        return
    pw_hash = hashlib.sha256(admin_password.encode("utf-8")).hexdigest()
    set_conf("root_password_sha2", pw_hash)
    set_conf("password_secret", password_secret)
    set_conf("is_master", str(is_master()).lower())

    api_port = get_api_port()
    if is_v2():
        if conf["web_listen_uri"]:
            url = urlparse(conf["web_listen_uri"])
            webport = db.get("web_listen_port")

            if webport == api_port:
                msg = (
                    "Config 'web_listen_uri' is using port {} which conflicts with"
                    " default REST port {}".format(webport, api_port)
                )
                hookenv.status_set("blocked", msg)
                return

            set_conf("web_listen_uri", conf["web_listen_uri"])
            if url.port != webport:
                if webport:
                    hookenv.close_port(webport)
                db.set("web_listen_port", url.port)
                hookenv.open_port(url.port)

        if conf["rest_transport_uri"]:
            set_conf("rest_transport_uri", validate_api_uri(conf["rest_transport_uri"]))
        else:
            clear_conf("rest_transport_uri")

        if conf["web_endpoint_uri"]:
            set_conf("web_endpoint_uri", validate_api_uri(conf["web_endpoint_uri"]))
        else:
            clear_conf("web_endpoint_uri")
    else:
        # We should not allow the configuration of `http_bind_address`
        # `http_bind_address` is equivalent to `rest_listen_uri` which defines
        # the IP and PORT used to liten for incoming connections. In Graylog 2,
        # we've never allowed the charm to change `rest_listen_uri`, so we
        # should not allow it change `http_bind_address` either
        # as the snap has 127.0.0.1:9000 as the default, we set it to
        # 0.0.0.0:9000 for consistency
        #
        #
        # Also, temporarily removing the ability to configure `http_publish_uri`,
        # This is a new config in Graylog 3, with no Graylog 2 equivalent. It
        # allows for overriding the IP published to the cluster discovery APIs,
        # which is used for other cluster nodes to communicate to this node.
        # A singular config value would be invalid, because several Graylog nodes
        # would all be boardcasting the same cluster API address.
        # TODO: LP#1866267 - Allow network look-ups for `http_publish_uri`
        # intead of guessing the best IP.
        set_conf("http_bind_address", "0.0.0.0:9000")
        set_conf(
            "http_publish_uri", validate_api_uri("http://0.0.0.0:{}/".format(api_port))
        )

        if conf["web_endpoint_uri"]:
            set_conf("http_external_uri", validate_api_uri(conf["web_endpoint_uri"]))
        else:
            clear_conf("http_external_uri")

    if conf.get("logextraction-rules"):
        rules_config = yaml.safe_load(conf.get("logextraction-rules"))
        if rules_config and isinstance(rules_config, list):
            _configure_logextraction(rules_config)

    set_jvm_heap_size(conf["jvm_heap_size"])

    hookenv.application_version_set(snap.get_installed_version(SNAP_NAME))
    hookenv.open_port(api_port)

    # Need to remove 'graylog_api.configured' flag here as a failsafe. Some config
    # changes do not require set_conf but will be handled by configure_graylog_api
    remove_state("graylog_api.configured")
    set_state("graylog.configured")
    report_status()


@when("leadership.is_leader")
@when_not_all("leadership.set.admin_password", "leadership.set.password_secret")
def generate_leader_passwords():
    """Generate the Graylog admin password if not shared via Juju config."""
    if not is_state("leadership.set.admin_password"):
        admin_password = host.pwgen(18)
        charms.leadership.leader_set(admin_password=admin_password)
    if not is_state("leadership.set.password_secret"):
        password_secret = host.pwgen(96)
        charms.leadership.leader_set(password_secret=password_secret)


@when_any("leadership.changed.admin_password", "leadership.changed.password_secret")
def get_leader_passwords():
    """Retrieve the stored admin password, if available."""
    db = unitdata.kv()
    admin_password = charms.leadership.leader_get("admin_password")
    db.set("admin_password", admin_password)
    password_secret = charms.leadership.leader_get("password_secret")
    db.set("password_secret", password_secret)
    remove_state("graylog.configured")


@when("leadership.is_leader")
def appoint_master():
    """Set the leader unit name for future operations."""
    master = charms.leadership.leader_get("master")
    if master and (
        master == hookenv.local_unit() or master in set(hookenv.expected_peer_units())
    ):
        # A master already exists
        return
    # Appoint myself as master
    charms.leadership.leader_set(master=hookenv.local_unit())


def is_master():  # noqa: D103
    return charms.leadership.leader_get("master") == hookenv.local_unit()


@when("leadership.changed.master")
@when("graylog.configured")
def set_master():
    """Store a boolean to track if the unit is the leader."""
    set_conf("is_master", str(is_master()).lower())


@when("graylog.configured")  # noqa: C901
def report_status():  # noqa: C901
    """Assess Graylog status via Juju status message."""
    beats_connected = is_state("beats.connected")
    beats_available = is_state("beat.setup")
    es_connected = is_state("elasticsearch.connected")
    es_available = is_state("elasticsearch.available")
    mongodb_connected = is_state("mongodb.connected")
    mongodb_available = is_state("mongodb.available")

    # Check for missing required applications
    missing_apps = []
    if not es_connected:
        missing_apps.append("elasticsearch")
    if not mongodb_connected:
        missing_apps.append("mongodb")
    channel_change_status = get_channel_change_status()
    if channel_change_status not in (
        ChannelChangeStatus.NO_CHANGE,
        ChannelChangeStatus.MINOR_CHANGE,
    ):
        if channel_change_status == ChannelChangeStatus.LATEST_REQUESTED:
            msg = (
                'The "latest" track is not supported; revert the channel config '
                "change to unblock"
            )
        elif channel_change_status == ChannelChangeStatus.INVALID:
            msg = "Invalid snap channel: {}".format(hookenv.config("channel"))
        elif channel_change_status == ChannelChangeStatus.UPGRADE_REQUESTED:
            msg = (
                "PACKAGE UPGRADE REQUIRES MANUAL INTERVENTION: see the charm's "
                "README.md for more info"
            )
        elif channel_change_status == ChannelChangeStatus.DOWNGRADE_REQUESTED:
            msg = (
                "PACKAGE DOWNGRADES ARE NOT SUPPORTED BY THIS CHARM - to clear this "
                "message, set config 'snap_channel' to '{}'".format(
                    snap.get_installed_channel(SNAP_NAME)
                )
            )
        else:
            raise ValueError(
                "Unexpected channel_block_type value: {!r}".format(
                    channel_change_status
                )
            )
        hookenv.status_set("blocked", msg)
    elif missing_apps:
        # Report blocked if we have missing apps
        msg = "Missing relation to: {}".format(", ".join(missing_apps))
        hookenv.status_set("blocked", msg)
    else:
        # Report ready/waiting based on current relation states
        waiting_apps = []
        ready_apps = []

        # Beats
        if beats_connected and not beats_available:
            waiting_apps.append("filebeat")
        elif beats_available:
            ready_apps.append("filebeat")

        # Elasticsearch
        if es_connected and not es_available:
            waiting_apps.append("elasticsearch")
        elif es_available:
            ready_apps.append("elasticsearch")

        # MongoDB
        if mongodb_connected and not mongodb_available:
            waiting_apps.append("mongodb")
        elif mongodb_available:
            ready_apps.append("mongodb")

        # Graylog REST API
        try:
            _verify_rest_api_is_alive(timeout=5)
        except ApiTimeout:
            waiting_apps.append("REST API")

        # Waiting apps take precedent over ready apps
        if waiting_apps:
            msg = "Waiting for: {}".format(", ".join(waiting_apps))
            hookenv.status_set("waiting", msg)
        else:
            # If no waiting_apps and we made it this far, we have ready_apps
            msg = "Ready with: {}".format(", ".join(ready_apps))
            hookenv.status_set("active", msg)


@when("graylog.configured")
@when("mongodb_config.set")
@when("elasticsearch_config.set")
@when_not("graylog.needs_restart")
@when_not("graylog_api.configured")
def configure_graylog_api(*discard):
    """Adjust states to ensure API dependents will be correctly configured."""
    try:
        _verify_rest_api_is_alive()
    except ApiTimeout:
        # Corner case: ES/Mongo are up, but REST API is not up yet.
        # Just wait (status already set in report_status()) and try again next time.
        pass
    else:
        remove_state("beat.setup")
        remove_state("graylog_index_sets.configured")
        remove_state("graylog_inputs.configured")
        remove_state("graylog_nagios.configured")
        set_state("graylog_api.configured")


@when("graylog.configured")  # noqa: C901
@when("graylog_api.configured")
@when_not("graylog_index_sets.configured")
def configure_index_sets(*discard):  # noqa: C901
    """Configure ES indices."""
    conf = hookenv.config()
    g = get_default_graylog_client()
    index_sets = g.index_set_get()
    if index_sets is None:
        return
    index_shards = conf["index_shards"]
    if index_shards == 0:
        # Automatically work out the index shards based on how
        # many Elasticsearch units.
        for relid in hookenv.relation_ids("elasticsearch"):
            index_shards += len(hookenv.related_units(relid))
        if not index_shards:
            hookenv.log("Can't work out number of elasticsearch units")
            return

    for iset in index_sets:
        log = ""
        if iset["shards"] != index_shards:
            log += " shards from {} to {}".format(iset["shards"], index_shards)
        iset["shards"] = index_shards
        if iset["replicas"] != conf["index_replicas"]:
            log += " replicas from {} to {}".format(
                iset["replicas"], conf["index_replicas"]
            )
        iset["replicas"] = conf["index_replicas"]
        if conf["index_rotation_strategy"] in rotation_strategies:
            rs = rotation_strategies[conf["index_rotation_strategy"]]
            if iset["rotation_strategy_class"] != rs["class"]:
                log += " rotation strategy from {} to {} ({}={})".format(
                    iset["rotation_strategy_class"],
                    rs["class"],
                    rs["special"],
                    conf[rs["juju-conf"]],
                )
            iset["rotation_strategy_class"] = rs["class"]
            iset["rotation_strategy"] = {
                "type": "{}Config".format(rs["class"]),
                rs["special"]: conf[rs["juju-conf"]],
            }
        index_retention_count = conf["index_retention_count"]
        if iset["retention_strategy"]["max_number_of_indices"] != index_retention_count:
            log += " retention max_number_of_indices={}".format(index_retention_count)
        iset[
            "retention_strategy_class"
        ] = "org.graylog2.indexer.retention.strategies.DeletionRetentionStrategy"
        iset["retention_strategy"] = {
            "type": (
                "org.graylog2.indexer.retention.strategies."
                "DeletionRetentionStrategyConfig"
            ),
            "max_number_of_indices": index_retention_count,
        }
        if not g.index_set_update(iset["id"], iset):
            return
        if log:
            hookenv.log("Updated{}".format(log))

    set_state("graylog_index_sets.configured")


def _configure_logextraction(rules_config):  # noqa: D103
    g = get_default_graylog_client()
    logext = LogExtractPipeline(g)
    logext.set_rules(rules_config)


def _check_input_exists(current_inputs, new):
    """Find if an input plugin is installed.

    Given the current list of configured Graylog inputs, check if new input
    exists and return an ID as well as "changed" flag to indicate that the
    new input provided has changed (input type, bind address, and port).
    """
    changed = True
    cur = None
    for input in current_inputs:
        if input["title"] == new["title"]:
            cur = input
            break
    if not cur:
        return (True, None)
    current = [
        cur["type"],
        cur["attributes"]["bind_address"],
        str(cur["attributes"]["port"]),
    ]
    new = [
        new["type"],
        new["configuration"]["bind_address"],
        str(new["configuration"]["port"]),
    ]
    if sorted(new) == sorted(current):
        changed = False
    return (changed, cur["id"])


def _close_uneeded_ports(current_inputs, new):
    for cur in current_inputs:
        title = cur["title"]
        # ignore Custom inputs from web interface
        if title.lower().startswith("custom"):
            continue
        # ignore juju_beats inputs from elastic-beats relation
        if title.lower() == "juju_beats_input":
            continue
        type = cur["type"]
        if "udp" in type.lower():
            proto = "UDP"
        else:
            proto = "TCP"
        port = cur["attributes"]["port"]
        if "{}/{}".format(port, proto) not in new:
            hookenv.close_port(port, proto)


def _remove_old_inputs(inputs, new_inputs):
    """Remove any old or previously defined input, or any unknown.

    Input plugins prefixed with 'Custom' will be kept to allow configuring additional
    inputs via the web UI.
    """
    g = get_default_graylog_client()
    success = True
    for cur in inputs:
        title = cur["title"]
        # ignore Custom inputs from web interface
        if title.lower().startswith("custom"):
            continue
        # ignore juju_beats inputs from elastic-beats relation
        if title.lower().startswith("juju_beats_"):
            continue
        if title in new_inputs:
            continue
        if g.log_input_remove(cur["id"]):
            hookenv.log("Removed old input: {} ({})".format(cur["title"], cur["id"]))
        else:
            success = False
    return success


@when("graylog_api.configured")
@when("beat.setup")
@when_not("beats.connected")
def graylog_remove_filebeat_input():
    """Remove the logstash beats client."""
    hookenv.log("Removing filebeats input")

    g = get_default_graylog_client()

    inputs = g.log_input_get()

    if inputs is not None:

        for inp in inputs:

            if inp["title"].lower() == "juju_beats_input":

                g.log_input_remove(inp["id"])
                hookenv.log(
                    "Removed beats input: {} ({})".format(inp["title"], inp["id"])
                )
                hookenv.status_set(
                    "active", "Removed beats input {}".format(inp["title"])
                )
    remove_state("beat.setup")


@when("beats.connected")
def provide_beats_port(filebeat):
    """Share beats input plugin destination via Juju relations."""
    beats_port = hookenv.config("beats_port")
    filebeat.provide_data(beats_port)


@when("graylog.configured")
@when("graylog_api.configured")
@when("beats.connected")
@when_not("beat.setup")
def setup_beats_input(filebeat):
    """Configure the beats input plugin."""
    port = hookenv.config("beats_port")
    hookenv.log("Setting up beats input on port {}".format(port))
    filebeat.provide_data(port)

    g = get_default_graylog_client()

    type = g.map_log_input_type("Beats")
    if type is None:
        return

    d = {
        "title": "juju_beats_input",
        "type": type,
        "global": "true",
        "configuration": {
            "bind_address": "0.0.0.0",
            "port": port,
            "tcp_keepalive": "true",
        },
    }

    proto = "TCP"
    hookenv.open_port(port, proto)
    inputs = g.log_input_get()

    (changed, input_id) = _check_input_exists(inputs, d)
    if changed:
        if input_id:
            hookenv.log("updating beats input with new port {}".format(port))
        else:
            hookenv.log("configuring beats input on port {}".format(port))
        ret = g.log_input_update(input_id, d)
        if not ret:
            return
    else:
        hookenv.log("beats input already configured")

    set_state("beat.setup")


@when("graylog.configured")
@when("graylog_api.configured")
@when_not("graylog_inputs.configured")
def configure_inputs(*discard):
    """Configure log inputs in Graylog via the API."""
    conf = hookenv.config()
    g = get_default_graylog_client()
    inputs = g.log_input_get()
    hookenv.log("Inputs: {}".format(inputs))
    if inputs is None:
        hookenv.log("Not configuring inputs.  GraylogApi returned None")
        return
    new_opened_ports = []
    new_inputs = []
    for new in yaml.safe_load(conf["log_inputs"]) or {}:
        type = g.map_log_input_type(new["type"])
        if not type:
            hookenv.log('Input type "{}" not supported'.format(new["type"]))
            continue
        d = {
            "title": new["name"],
            "type": type,
            "global": "true",
            "configuration": {
                "bind_address": new["bind_address"],
                "port": new["bind_port"],
            },
        }
        if "udp" in type.lower():
            proto = "UDP"
        else:
            proto = "TCP"
            d["configuration"]["tcp_keepalive"] = "true"

        hookenv.open_port(new["bind_port"], proto)
        new_opened_ports.append("{}/{}".format(new["bind_port"], proto))

        (changed, input_id) = _check_input_exists(inputs, d)
        if not changed:
            new_inputs.append(new["name"])
            continue
        if input_id:
            change_text = "Updated existing input"
        else:
            change_text = "Adding new input"
        ret = g.log_input_update(input_id, d)
        if not ret:
            # Can't add input, let's try again later
            return
        hookenv.log("{}: {} ({})".format(change_text, new["name"], ret["id"]))
        new_inputs.append(new["name"])

    # Now remove inputs that should no longer be there as well as close ports
    _close_uneeded_ports(inputs, new_opened_ports)
    refreshed_inputs = g.log_input_get()
    if not _remove_old_inputs(refreshed_inputs, new_inputs):
        # Can't delete input, let's try again later
        return
    set_state("graylog_inputs.configured")


@when("graylog.configured")
@when("elasticsearch.available")
def configure_elasticsearch_connection(elasticsearch):
    """Configure ES parameters in Graylog's configuration file."""
    cluster = ""
    # Pre-2.3.x discovery unicast hosts (binary ElasticSearch protocol)
    discovery_hosts = []
    # 2.3.x and above HTTP REST API
    http_hosts = []
    for unit in elasticsearch.list_unit_data():
        cluster_name = unit["cluster_name"]
        if cluster_name is not None:
            cluster = cluster_name
        http_hosts.append("http://{}:{}".format(unit["host"], unit["port"]))
        discovery_hosts.append(
            "{}:{}".format(unit["host"], ELASTICSEARCH_DISCOVERY_PORT)
        )

    if not cluster:
        conf = hookenv.config()
        cluster = conf["elasticsearch_cluster_name"]

    if not data_changed(
        "elasticsearch.relation", {"cluster_name": cluster, "hosts": http_hosts}
    ):
        return

    set_conf("elasticsearch_cluster_name", cluster)
    set_conf(
        "elasticsearch_discovery_zen_ping_unicast_hosts ", ", ".join(discovery_hosts)
    )
    set_conf("elasticsearch_hosts", ", ".join(http_hosts))
    # Elastic search does not reliably pick the right ip to listen on
    set_conf("elasticsearch_network_host", hookenv.unit_private_ip())
    set_state("elasticsearch_config.set")

    hookenv.log("Updated elasticsearch configuration")
    report_status()


@when("graylog.configured")
@when("mongodb.available")
def configure_mongodb_connection(mongodb):
    """Configure MongoDB parameters in Graylog's configuration file."""
    mongodb_hosts = []

    for mongo_host in mongodb.connection_strings():
        mongodb_hosts.append(mongo_host)

    mongodb_uri = "mongodb://{}/graylog".format(",".join(mongodb_hosts))
    if not data_changed("mongodb.uri", mongodb_uri):
        return

    set_conf("mongodb_uri", mongodb_uri)
    set_state("mongodb_config.set")
    hookenv.log("Updated mongodb configuration")
    report_status()


@when("graylog.configured")
@when("website.available")
def update_reverseproxy_config(website):
    """Share Graylog endpoint details for a reverse proxy to use."""
    services = ""
    conf = hookenv.config()
    if conf["web_listen_uri"]:
        url = urlparse(conf["web_listen_uri"])
        website.configure(port=url.port)
        services += "- {service_name: web, service_port: " + str(url.port) + "}\n"
    if is_v2():
        # NB: there is no revproxy to the api endpoint in v3+.
        services += "- {service_name: api, service_port: " + get_api_port() + "}\n"
    website.set_remote(all_services=services)


@when("graylog.configured")
@when("graylog.needs_restart")
def restart_service(service=SERVICE_NAME):
    """Restart the service and on failure wait 15 seconds and try again.

    This handles situations when relations are adding quick enough to trigger
    systemd's standard protection against frequent restarts.
    """
    host.service_restart(service)
    remove_state("graylog.needs_restart")
    if host.service_running(service):
        return

    time.sleep(15)
    host.service_restart(service)


def set_conf(key, value, conf_path=CONF_FILE):
    """Update the graylog configuration setting lines matching 'key = value'.

    The configuration contains randomly generated values specific to each install
    so setting the configuration this way avoids having to regenerate those values.
    """
    # keys, values should only be strings as they are intended
    # to be used directly in the graylog configuration file (text)
    assert isinstance(key, str), "key should be a string"
    assert isinstance(value, str), "value should be a string"

    conf = []
    key_re = re.compile("^#*\\s*{}".format(key))

    changed = False
    replaced = False
    with open(conf_path, "rb") as conf_file:
        found = False
        for line in conf_file:
            line = line.decode().rstrip()
            if not key_re.match(line):
                conf.append(line)
                continue
            found = True
            if not replaced:
                if value != "":  # empty strings are removed, update the rest
                    new = "{} = {}".format(key, value).rstrip()
                    if line != new:
                        hookenv.log(
                            'Updating configuration file option "{}".'.format(key)
                        )
                        changed = True
                    conf.append(new)
                else:
                    # Allow for a config to be cleared from the config file by setting
                    # to ""
                    if not line.startswith("#"):
                        hookenv.log(
                            'Removing configuration file option "{}".'.format(key)
                        )
                        conf.append("#" + line)
                        changed = True
                    else:
                        conf.append(line)
                replaced = True
            elif not line.startswith("#"):
                conf.append("#" + line)
            else:
                conf.append(line)
        if not found:
            hookenv.log('Updating configuration file new option "{}".'.format(key))
            changed = True
            conf.append("{} = {}".format(key, value))

    if changed:
        flag_restart_and_api_reconfigure_needed()
        with open(conf_path, "wb") as conf_file:
            conf_file.write("\n".join(conf).encode())
        return True

    return False


def clear_conf(key, conf_path=CONF_FILE):
    """Call set_conf with an empty string.

    Comments out the given field append allow Graylog to select its own defaults.
    """
    return set_conf(key, "", conf_path)


def parse_java_opts(opts, size):  # noqa: D103
    found_xms = False
    found_xmx = False
    new_opts = []
    for opt in opts.split():
        if opt.startswith("-Xms"):
            found_xms = True
            if size == "":
                continue
            new_opts.append("-Xms{}".format(size))
        elif opt.startswith("-Xmx"):
            found_xmx = True
            if size == "":
                continue
            new_opts.append("-Xmx{}".format(size))
        else:
            new_opts.append(opt)
    if not found_xms:
        new_opts.append("-Xms{}".format(size))
    if not found_xmx:
        new_opts.append("-Xmx{}".format(size))
    return new_opts


def set_jvm_heap_size(heap_size="1G", conf_path=SERVER_DEFAULT_CONF_FILE):  # noqa: D103
    size = heap_size.lower()
    conf = []
    key_re = re.compile('^\\s*GRAYLOG_SERVER_JAVA_OPTS\\s*="(.*)"$')
    changed = False
    # If specified /etc/default/graylog-server doesn't exist, use the default
    # file shipped out by the snap
    if not os.path.exists(conf_path):
        updated_conf_path = SHIPPED_SNAP_SERVER_DEFAULT_CONF_FILE
    else:
        updated_conf_path = conf_path
    with open(updated_conf_path, "rb") as conf_file:
        for line in conf_file:
            line = line.decode()
            m = key_re.match(line)
            if not m:
                conf.append(line)
                continue
            new_opts = parse_java_opts(m.group(1), size)
            new_line = 'GRAYLOG_SERVER_JAVA_OPTS="{}"\n'.format(" ".join(new_opts))
            if new_line != line:
                hookenv.log(
                    'Updating GRAYLOG_SERVER_JAVA_OPTS="{}".'.format("".join(new_opts))
                )
                changed = True
            conf.append(new_line)

    if changed:
        flag_restart_and_api_reconfigure_needed()
        with open(conf_path, "wb") as conf_file:
            conf_file.write("".join(conf).encode())
        return True
    return False


def flag_restart_and_api_reconfigure_needed():  # noqa: D103
    set_state("graylog.needs_restart")

    # Re-run API configuration in case of relevant config changes.
    # Note that config can be changed via charm config, relations, and in
    # other ways.  If we need a restart, we should assume that we need to re-run
    # our API configuration.
    #
    # Important: charms.reactive does not re-evaluate queued handlers when
    # flags get set, but it does when flags get cleared.  Thus, the above is not
    # sufficient to cause the conditions on queued handlers to be re-evaluated.
    # Thus, rather than clearing this flag as a side-effect of the
    # restart_service() hanlder, the flag is cleared here to ensure queued
    # handlers with graylog.needs_restart dependencies don't accidentally run.

    remove_state("graylog_api.configured")


@when("leadership.is_leader")
@when("nrpe-external-master.available")
@when_not("leadership.set.nagios_password")
def set_up_nagios_user():
    """Configure a user for Nagios to use."""
    nagios_password = host.pwgen(18)

    db = unitdata.kv()
    db.set("nagios_password", nagios_password)

    charms.leadership.leader_set(nagios_password=nagios_password)


@when("leadership.is_leader")
@when("leadership.set.nagios_password")
@when_not("leadership.set.nagios_token")
def set_up_nagios_token():
    """Configure a token for Nagios to use."""
    # For nagios checks via API, we need to create a user and token
    db = unitdata.kv()
    nagios_password = db.get("nagios_password")

    api_url = get_api_url()
    g = get_default_graylog_client()
    g.user_create(NAGIOS_USERNAME, nagios_password)
    g.user_permissions_set(
        NAGIOS_USERNAME,
        [
            "users:tokenlist",
            "users:tokencreate",
            "indexercluster:read",
            "indices:failures",
            "notifications:read",
            "journal:read",
        ],
    )
    g = GraylogApi(
        base_url=api_url,
        username=NAGIOS_USERNAME,
        password=nagios_password,
        token_name=NAGIOS_USERNAME,
    )
    nagios_token = g.token_get()
    if not nagios_token:
        hookenv.log("No nagios token, return from setup early")
        return

    charms.leadership.leader_set(nagios_token=nagios_token)


@when_any("leadership.changed.nagios_password", "leadership.changed.nagios_token")
def set_nagios_credentials():
    """Store Nagios credentials for later use."""
    db = unitdata.kv()
    nagios_password = charms.leadership.leader_get("nagios_password")
    db.set("nagios_password", nagios_password)
    nagios_token = charms.leadership.leader_get("nagios_token")
    db.set("nagios_token", nagios_token)


@when("graylog.configured")
@when("graylog_api.configured")
@when("nrpe-external-master.available")
@when_not("graylog_nagios.configured")
def configure_nagios(nagios):
    """Configure Nagios requirements for nrpe check to work."""
    db = unitdata.kv()
    nagios_password = db.get("nagios_password")
    nagios_token = db.get("nagios_token")
    if not (nagios_password and nagios_token):
        # we wait for the leader to set up the nagios user
        hookenv.status_set("maintenance", "Waiting for leader to set up nagios")
        return

    api_port = get_api_port()
    api_url = get_api_url()

    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname, primary=True)

    conf = hookenv.config()
    if conf["web_listen_uri"]:
        url = urlparse(conf["web_listen_uri"])
        # LP:#1706265: Work around this bug by escaping some characters.
        check_string = "\\<title\\>Graylog\\ Web\\ Interface\\</title\\>"
        url_path = url.path
        # Catch URI without an explicit path
        if not url_path:
            url_path = "/"
        # Ensure an ending slash
        if not url_path.endswith("/"):
            url_path += "/"
        nrpe_setup.add_check(
            "graylog_http",
            "Graylog Web UI check",
            '/usr/lib/nagios/plugins/check_http -I {} -p {} -u "{}" -s "{}"'.format(
                url.hostname, str(url.port), url_path, check_string
            ),
        )
    check_string = "cluster_id"
    nrpe_setup.add_check(
        "graylog_api",
        "Greylog API check",
        (
            '/usr/lib/nagios/plugins/check_http -k "Accept:\\ application/json" '
            '-I 127.0.0.1 -p {} -u /api -s "{}"'
        ).format(api_port, check_string),
    )

    host.write_file(
        "/var/lib/nagios/graylog-api-token",
        nagios_token.encode(),
        group="nagios",
        perms=0o640,
    )
    nagios_plugins = "/usr/local/lib/nagios/plugins/"
    for f in ("files/check_graylog_health.py", "lib/charms/layer/graylog/api.py"):
        with open(f, "rb") as content:
            fname = os.path.join(nagios_plugins, os.path.basename(f))
            host.write_file(fname, content.read(), perms=0o755)
    nrpe_setup.add_check(
        "graylog_health",
        "Graylog Health check",
        "{} --uncommitted-warning {} --uncommitted-critical {} {}".format(
            os.path.join(nagios_plugins, "check_graylog_health.py"),
            conf.get("nagios_uncommitted_warn", 0),
            conf.get("nagios_uncommitted_crit", 100),
            api_url,
        ),
    )

    nrpe_setup.write()
    set_state("graylog_nagios.configured")


def get_default_graylog_client():  # noqa: D103
    db = unitdata.kv()
    return GraylogApi(
        base_url=get_api_url(),
        username="admin",
        password=db.get("admin_password"),
        token_name="graylog-charm",
    )


def _verify_rest_api_is_alive(timeout=DEFAULT_REST_API_TIMEOUT):
    hookenv.log("Verifying REST API is alive...")
    g = get_default_graylog_client()
    url = ""  # Will query using the base URL of the client, i.e. /api/
    resp = g.request(url)
    start_ts = time.time()
    while resp is None:
        time.sleep(5)
        hookenv.log("Retrying REST API check...")
        resp = g.request(url)
        if time.time() - start_ts > timeout:
            raise ApiTimeout()
    hookenv.log("REST API is up")
