"""Encapsulate graylog testing."""

import logging
import time
import unittest

from api import GraylogApi

import tenacity

import zaza.model as model


CURL_TIMEOUT = 180
REQ_TIMEOUT = 12
DEFAULT_API_PORT = "9001"
DEFAULT_WEB_PORT = "9000"


class BaseGraylogTest(unittest.TestCase):
    """Base for Graylog charm tests."""

    @classmethod
    def setUpClass(cls):
        """Initialize Graylog default values by API version."""
        super(BaseGraylogTest, cls).setUpClass()
        cls.model_name = model.get_juju_model()
        cls.application_name = "graylog"
        cls.lead_unit_name = model.get_lead_unit_name(
            cls.application_name, model_name=cls.model_name
        )
        cls.units = model.get_units(cls.application_name, model_name=cls.model_name)
        logging.debug("Leader unit is {}".format(cls.lead_unit_name))
        cls.graylog_ip = model.get_app_ips(cls.application_name)[0]

        # get the graylog version from the configuration
        cls.config = model.get_application_config(cls.application_name)
        channel = cls.config.get("channel").get("value")
        logging.debug("Snap channel is: {}".format(channel))

        # account for the differences between graylog2 and 3
        if channel.startswith("2"):
            cls.api_port = DEFAULT_API_PORT
            cls.file_field = "file"
            cls.expected_plugins = {
                "org.graylog.plugins.pipelineprocessor.ProcessorPlugin",
                "org.graylog.plugins.beats.BeatsInputPlugin",
                "org.graylog.plugins.collector.CollectorPlugin",
            }
        else:  # graylog 3+
            cls.api_port = DEFAULT_WEB_PORT
            cls.file_field = "filebeat_log_file_path"
            cls.expected_plugins = {
                "org.graylog.aws.AWSPlugin",
                "org.graylog.plugins.collector.CollectorPlugin",
                "org.graylog.plugins.threatintel.ThreatIntelPlugin",
            }

        api_url = "http://{}:{}/api".format(cls.graylog_ip, cls.api_port)
        action = model.run_action_on_leader(cls.application_name, "show-admin-password")
        res = action.data["results"]
        cls.api = GraylogApi(api_url, "admin", res["admin-password"])
        # try harder to get an api connection to cater for overloaded testsystems
        cls.api.req_timeout = REQ_TIMEOUT
        logging.debug("API at {}".format(api_url))


class CharmOperationTest(BaseGraylogTest):
    """Verify operations."""

    def test_05_api_ready(self):
        """Test if the API is ready.

        Curl the api endpoint on the sentried graylog unit. We'll retry
        until the CURL_TIMEOUT because it may take a few seconds for the
        graylog systemd service to start.
        """
        curl_command = "curl http://localhost:{}".format(self.api_port)
        timeout = time.time() + CURL_TIMEOUT
        while time.time() < timeout:
            response = model.run_on_unit(self.lead_unit_name, curl_command)
            if response["Code"] == "0":
                return
            logging.warning(
                "Unexpected curl response: {}. Retrying in 30s.".format(response)
            )
            time.sleep(30)

        # we didn't get rc=0 in the alloted time, fail the test
        self.fail(
            "Graylog didn't respond to the command \n"
            "'{curl_command}' as expected.\n"
            "Result: {result}".format(curl_command=curl_command, result=response)
        )

    def test_06_elasticsearch_healthy(self):
        """Verify if the Elasticsearch cluster is healthy (green)."""
        resp = self.api.indexer_cluster_health()
        self.assertEqual(resp["status"], "green")

    def test_07_elasticsearch_no_faults(self):
        """Verify if indexer failures were raised."""
        resp = self.api.indexer_failures_count("2020-01-01")
        self.assertEqual(int(resp["count"]), 0)

    def test_08_plugins(self):
        """Verify if extra plugins were created after Juju relations were applied."""
        resp = self.api.request("/system/plugins")
        plugin_set = set([p["unique_id"] for p in resp["plugins"]])

        self.assertTrue(self.expected_plugins <= plugin_set)

    def test_09_cluster(self):
        """Verify Graylog cluster health."""
        resp = self.api.cluster_get()
        self.assertGreater(len(resp.values()), 0)
        node = list(resp.values()).pop()
        self.assertTrue(node["is_processing"])
        self.assertEqual(node["lifecycle"], "running")

    def test_10_keyword_search(self):
        """Test searching a keyword on a logfile."""
        params = {
            "query": (
                '{}: "/var/log/cloud-init-output.log"'
                " AND ci-info".format(self.file_field)
            ),
            "keyword": "4 hours ago",
        }
        for attempt in tenacity.Retrying(
            wait=tenacity.wait_fixed(10),  # seconds
            stop=tenacity.stop_after_attempt(3),
            reraise=True,
        ):
            with attempt:
                resp = self.api.request("/search/universal/keyword", params=params)
                self.assertTrue(int(resp["total_results"]) > 0)
