"""Tests for graylog version upgrades."""

import time
import unittest

from zaza import model


class UpgradeTest(unittest.TestCase):
    """Tests for graylog version upgrades via snap channel changes."""

    def setUp(self):
        """Test setup."""
        self.unit_names = [unit.name for unit in model.get_units("graylog")]

    def test_01_channel_changed_to_newer_and_back(self):
        """Verify that we can change channel "up" and back.

        Changing the channel to a newer version should just result in a blocked state,
        until an action is manually run to finalize the change. Thus, we should be able
        to change to the new channel in the config, and as long as we don't run that
        action, also revert our change and get back to a sane state.

        """
        # Assumption: graylog channel is 2/stable before reaching this test
        self.assertEqual(
            model.get_application_config("graylog")["channel"].get("value"), "2/stable"
        )

        # Upon attempting an upgrade, we should go to a blocked state, with a reference
        # to instructions on how to complete the upgrade.
        model.set_application_config("graylog", {"channel": "3/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name,
                "blocked",
                "PACKAGE UPGRADE REQUIRES MANUAL INTERVENTION: see the charm's"
                " README.md for more info",
            )

        # Let's roll back to 2/stable; at this point it's still supported
        model.set_application_config("graylog", {"channel": "2/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "active", "Ready with: filebeat, elasticsearch, mongodb"
            )

    def test_02_channel_changed_to_newer_and_finalized(self):
        """Verify that we can change channel "up" and commit the change.

        After committing the change, we cannot roll back, as we'll see in later tests.

        """
        # Assumption: graylog channel is 2/stable before reaching this test
        self.assertEqual(
            model.get_application_config("graylog")["channel"].get("value"), "2/stable"
        )

        # Upon attempting an upgrade, we should go to a blocked state, with a reference
        # to instructions on how to complete the upgrade.
        model.set_application_config("graylog", {"channel": "3/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name,
                "blocked",
                "PACKAGE UPGRADE REQUIRES MANUAL INTERVENTION: see the charm's"
                " README.md for more info",
            )

        # Perform the "manual" upgrade
        for unit_name in self.unit_names:
            model.run_action(unit_name, "do-upgrade")

        # Verify that we've come back out the other side in one piece.
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "active", "Ready with: filebeat, elasticsearch, mongodb"
            )

    def test_03_channel_changed_to_older_and_back(self):
        """Verify that we can change channel "down" and back.

        Changing the channel to an older version is blocked and not allowed. Restoring
        the previous setting should remove that blocked state.

        """
        # Assumption: graylog channel is 3/stable before reaching this test
        self.assertEqual(
            model.get_application_config("graylog")["channel"].get("value"), "3/stable"
        )

        # Upon attempting a downgrade, we should go to a blocked state, explaining that
        # downgrades are not supported.
        # (Reverting the config setting back will unblock the app.)
        model.set_application_config("graylog", {"channel": "2/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name,
                "blocked",
                "PACKAGE DOWNGRADES ARE NOT SUPPORTED BY THIS CHARM - to clear "
                "this message, set config 'snap_channel' to '3/stable'",
            )

        # Let's roll back to 2/stable; at this point it's still supported
        model.set_application_config("graylog", {"channel": "3/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "active", "Ready with: filebeat, elasticsearch, mongodb"
            )

    def test_04_channel_changed_to_latest_and_back(self):
        """Verify behavior when changing to channel "latest"."""
        # Assumption: graylog channel is 3/stable before reaching this test
        self.assertEqual(
            model.get_application_config("graylog")["channel"].get("value"), "3/stable"
        )

        # We shouldn't allow upgrading to anything in the "latest" track; this may be
        # newer *or* older than the current channel.
        # Require an explicit track, i.e. 2/* or 3/*..
        model.set_application_config("graylog", {"channel": "latest/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name,
                "blocked",
                'The "latest" track is not supported; revert the channel config change'
                " to unblock",
            )

        # Verify that we've come back out the other side in one piece.
        model.set_application_config("graylog", {"channel": "3/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "active", "Ready with: filebeat, elasticsearch, mongodb"
            )

    def test_05_channel_changed_to_invalid_value(self):
        """Verify behavior when changing to an invalid channel."""
        # Assumption: graylog channel is 3/stable before reaching this test
        self.assertEqual(
            model.get_application_config("graylog")["channel"].get("value"), "3/stable"
        )

        # The channel can be set to an arbitrary value; we should catch if the
        # channel is bogus.
        model.set_application_config("graylog", {"channel": "fake"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "blocked", "Invalid snap channel: fake",
            )

        # Roll back
        model.set_application_config("graylog", {"channel": "3/stable"})
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "active", "Ready with: filebeat, elasticsearch, mongodb"
            )

    def test_06_channel_changed_within_track(self):
        """Verify behavior when changing to a different channel but the same "track"."""
        # Assumption: graylog channel is 3/stable before reaching this test
        self.assertEqual(
            model.get_application_config("graylog")["channel"].get("value"), "3/stable"
        )

        # Snap refreshes within the same major series should be allowed.
        model.set_application_config("graylog", {"channel": "3/edge"})
        # config changed may not run right away; wait a bit, else we might see the
        # pre-update "active" state and think we're done.
        time.sleep(10)
        for unit_name in self.unit_names:
            self._block_until(
                unit_name, "active", "Ready with: filebeat, elasticsearch, mongodb"
            )

    def _block_until(
        self, unit, expected_workload_status, expected_workload_status_message
    ):
        app = unit.split("/")[0]
        # Zaza doesn't presently seem to provide a way to check these 3 things
        # at the same time.  I could write a helper to do this, but this may be "good
        # enough" for the time being.
        model.block_until_unit_wl_status(unit, expected_workload_status, timeout=300)
        model.block_until_wl_status_info_starts_with(
            app, expected_workload_status_message, timeout=300
        )
        model.wait_for_unit_idle(unit, timeout=300)
