"""Graylog v2 legacy tests."""
import re
import unittest

from api import GraylogApi

import yaml

from zaza import model
from zaza.utilities import juju


DEFAULT_API_PORT = "9001"  # Graylog 2 only
DEFAULT_WEB_PORT = "9000"


class LegacyTests(unittest.TestCase):
    """These tests were ported from tests/test_10_basic.py in changeset a3b54c2.

    They were temporarily deleted during the conversion from Amulet to Zaza.
    Aside from porting to work with libjuju and some refactoring, these tests
    are largely as they were before.
    """

    def test_api_ready(self):
        """Curl the api endpoint on the graylog unit."""
        port = self.get_api_port()
        curl_command = "curl http://localhost:{} --connect-timeout 30".format(port)
        self.run_command("graylog/0", curl_command)

    def test_elasticsearch_active(self):
        """Verify if the Elasticsearch cluster is healthy (green)."""
        self._assert_remote_address_in_server_conf("elastic/0", "client")
        # This was also stuck in the same test...
        api = self.get_graylog_clients()[0]
        resp = api.indexer_cluster_health()
        self.assertTrue(resp["status"] == "green")

    def test_mongodb_active(self):
        """Verify if the mongodb backend is healthy."""
        self._assert_remote_address_in_server_conf("mongo/0", "database")
        # This was also stuck in the same test...
        api = self.get_graylog_clients()[0]
        resp = api.cluster_get()
        self.assertTrue(resp[list(resp)[0]]["is_processing"])

    def _assert_remote_address_in_server_conf(self, remote_unit, remote_relation_name):
        """Look for the remote IP address in Graylog's configuration file."""
        # This is a very simplistic, naive check, but it's equivalent to
        # that from tests/test_10_basic.py in a3b54c2.
        source_unit = "graylog/0"
        remote_addr = juju.get_relation_from_unit(
            source_unit, remote_unit, remote_relation_name
        )["private-address"]
        assert remote_addr in self.get_file_contents(
            source_unit, "/var/snap/graylog/common/server.conf"
        )

    def get_graylog_clients(self):
        """Return the list of API clients to Graylog units."""
        graylog_units = juju.get_full_juju_status().applications["graylog"]["units"]
        graylog_addrs = [unit["public-address"] for unit in graylog_units.values()]
        port = self.get_api_port()
        password = self.get_graylog_admin_password()
        return [
            GraylogApi("http://{}:{}/api".format(graylog_addr, port), "admin", password)
            for graylog_addr in graylog_addrs
        ]

    def get_api_port(self):  # noqa: D102
        port = DEFAULT_API_PORT if self.is_v2() else DEFAULT_WEB_PORT
        return port

    def get_graylog_admin_password(self):  # noqa: D102
        result = model.run_action("graylog/0", "show-admin-password")
        password = result.data["results"]["admin-password"]
        return password

    def test_website_active(self):
        """Verify if the Graylog endpoints are correctly configured."""
        data = juju.get_relation_from_unit("haproxy/0", "graylog/0", "website")
        self.assertEqual(data["port"], DEFAULT_WEB_PORT)

        service_ports = {}
        for service in yaml.safe_load(data["all_services"]):
            service_ports[service["service_name"]] = service["service_port"]

        self.assertEqual(str(service_ports["web"]), DEFAULT_WEB_PORT)
        if self.is_v2():
            self.assertEqual(str(service_ports["api"]), DEFAULT_API_PORT)

    def run_command(self, target, command, timeout=30):
        """Run commands on remote units."""
        result = model.run_on_unit(target, command, timeout=timeout)
        assert result["Code"] == "0"
        return result

    def is_v2(self):
        """Validate Graylog API version (expect v2)."""
        return (
            juju.get_full_juju_status()
            .applications["graylog"]["workload-version"]
            .startswith("2.")
        )

    def test_nrpe_config(self):
        """Verify that the specific nrpe check is configured."""
        cfg = self.get_file_contents(
            "nrpe/0", "/etc/nagios/nrpe.d/check_graylog_health.cfg"
        )
        self.assertTrue(re.search(r"command.*check_graylog_health", cfg))

    def test_nrpe_health_check(self):
        """Verify that the nrpe script is available to be run by nagios-nrpe-server."""
        cfg = self.get_file_contents(
            "nrpe/0", "/usr/local/lib/nagios/plugins/check_graylog_health.py"
        )
        self.assertTrue(re.search(r"CRITICAL", cfg))

    def get_file_contents(self, unit, filename):
        """Retrieve content of a file in a remote unit."""
        result = self.run_command(unit, "cat '{}'".format(filename))
        assert result["Code"] == "0"
        return result["Stdout"]
